import shutil

from polyrename.transformation import TRANSFORMATIONS


def main(files):
    print("Choose a transformation:")
    for i, tranformation in enumerate(TRANSFORMATIONS):
        print("{}) {}".format(i, tranformation.schema["metadata"]["name"]))

    # Ask for selection input
    choice = None
    while choice not in range(len(TRANSFORMATIONS)):
        try:
            choice = int(input("Selection: "))
        except ValueError:
            pass

    # Gather arguments for selected transformation
    transformation_class = TRANSFORMATIONS[choice]
    transformation_args = []
    for option in transformation_class.schema["options"]:

        # Change prompt to show default value if it is specified
        if option["required"] and "default_value" in option:
            input_prompt = "{} ({}) (Default: {}):".format(
                option["name"], option["description"], option["default_value"]
            )
        else:
            input_prompt = "{} ({}):".format(option["name"], option["description"])

        choice = None
        while choice is None:
            try:
                user_input = input(input_prompt)

                # Allow for none_value to be used
                if user_input != "":
                    choice = option["datatype"](user_input)
                else:
                    if "default_value" in option:
                        choice = option["default_value"]
            except ValueError:
                print("Invalid input!")
        transformation_args.append(choice)

    # Create transformation with selected arguments
    tranformation = transformation_class(*transformation_args)
    transformed_file_sequence = tranformation.resolve(files)
    print("Preview:")
    file_pairs = list(zip(files, transformed_file_sequence))
    for file_pair in file_pairs:
        print("{} -> {}".format(*file_pair))

    # Ask to perform rename operation
    response = None
    while response not in ["y", "n"]:
        response = input("Perform rename operation(s)? (y/n): ").lower()

    if response == "y":
        for file_pair in file_pairs:
            shutil.move(*file_pair)



from PySide2.QtWidgets import (
    QMainWindow,
    QWidget,
    QDesktopWidget,
    QGridLayout,
    QAction,
    QMessageBox,
)
from PySide2.QtGui import QIcon
from PySide2.QtCore import QSize

from polyrename.driver.gui import ICON_ROOT
from polyrename.driver.gui.file_picker import FilePicker
from polyrename.driver.gui.transformation_library import TransformationLibrary
from polyrename.driver.gui.transformation_configuration import (
    TransformationConfiguration,
)
from polyrename.driver.gui.pipeline_editor import PipelineEditor


class MainWindow(QMainWindow):
    def __init__(self, files):
        super().__init__()

        self._init_window()
        self._init_logo()
        self._init_layout()

        # Initialize each quadrant
        self.file_picker = FilePicker(files)
        self.grid_layout.addWidget(self.file_picker, 0, 1)

        self.pipeline_editor = PipelineEditor(self.file_picker)
        self.grid_layout.addWidget(self.pipeline_editor, 0, 0)

        self.transformation_configuration = TransformationConfiguration(
            self.pipeline_editor
        )
        self.grid_layout.addWidget(self.transformation_configuration, 1, 1)

        self.transformation_library = TransformationLibrary(
            self.transformation_configuration
        )
        self.grid_layout.addWidget(self.transformation_library, 1, 0)

        self.init_menu_bar()

        self.show()

    def _init_window(self):
        """Initialize main window size and position"""

        self.setWindowTitle("PolyRename")
        self.resize(1280, 720)

        # Center window on screen
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

    def _init_logo(self):
        """Initialize application icon"""

        app_icon = QIcon()
        app_icon.addFile(str(ICON_ROOT / "16x16.png"), QSize(16, 16))
        app_icon.addFile(str(ICON_ROOT / "32x32.png"), QSize(32, 32))
        app_icon.addFile(str(ICON_ROOT / "64x64.png"), QSize(64, 64))
        app_icon.addFile(str(ICON_ROOT / "128x128.png"), QSize(128, 128))
        app_icon.addFile(str(ICON_ROOT / "256x256.png"), QSize(256, 256))

        self.setWindowIcon(app_icon)

    def _init_layout(self):
        """Initialize base layout of main windows"""

        self.grid_layout = QGridLayout()
        self.grid_layout.setSpacing(10)

        central_widget = QWidget()
        central_widget.setLayout(self.grid_layout)
        self.setCentralWidget(central_widget)

    def init_menu_bar(self):
        """Initialize Menu Bar and contained menus"""

        menu_bar = self.menuBar()
        file_menu = menu_bar.addMenu("&File")
        # edit_menu = menu_bar.addMenu("&Edit")
        help_menu = menu_bar.addMenu("&Help")

        select_file_action = QAction("S&elect Files", self)
        select_file_action.setShortcut("Ctrl+e")
        select_file_action.setStatusTip("Select Files")
        select_file_action.triggered.connect(self.file_picker._select_files_listener)
        file_menu.addAction(select_file_action)

        quit_action = QAction("&Quit", self)
        quit_action.setShortcut("Ctrl+q")
        quit_action.setStatusTip("Quit PolyRename")
        quit_action.triggered.connect(self.close)
        file_menu.addAction(quit_action)

        # undo_action = QAction("&Undo", self)
        # undo_action.setShortcut("Ctrl+Z")
        # undo_action.setStatusTip("Undo last action")
        # edit_menu.addAction(undo_action)
        #
        # redo_action = QAction("&Redo", self)
        # redo_action.setShortcut("Ctrl+Shift+Z")
        # redo_action.setStatusTip("Redo last action")
        # edit_menu.addAction(redo_action)

        about_action = QAction("&About", self)
        about_action.setShortcut("Ctrl+?")
        about_action.setStatusTip("About PolyRename")
        about_action.triggered.connect(self._about_handler)
        help_menu.addAction(about_action)

        # Initialize Status Bar
        status_bar = self.statusBar()
        status_bar.showMessage("Ready")

    def _about_handler(self):
        info = (
            "PolyRename is a cross-platform bulk-rename tool.\n"
            "\n"
            "Created By. Andrew Simmons & Ethan Smith\n"
        )
        about_dialog = QMessageBox.about(self, "About PolyRename", info)



def insert_text_before_extension(file, insert_text):
    """Inserts text before the extension of a path"""

    file_name = file.stem + insert_text + file.suffix
    file_path = file.parent / file_name
    return file_path



from polyrename.transformation.transformation import Transformation
from polyrename.transformation.utils.path_utils import insert_text_before_extension


class InsertTransformation(Transformation):
    schema = {
        "metadata": {
            "name": "Insert",
            "description": "Inserts text at specified location",
        },
        "options": [
            {
                "name": "Text",
                "description": "Text to insert",
                "datatype": str,
                "required": True,
            },
            {
                "name": "Position",
                "description": "Position to insert text at",
                "datatype": int,
                "required": True,
            },
        ],
    }

    def __init__(self, text, position):
        self.text = text
        self.position = position

    def resolve(self, file_sequence):
        return_sequence = []
        for file in file_sequence:

            if len(file.stem) < self.position:
                file = insert_text_before_extension(file, self.text)
            else:
                before = file.stem[0 : self.position]
                after = file.stem[self.position :]
                file = file.parent / (before + self.text + after + file.suffix)

            return_sequence.append(file)

        return return_sequence

    def __repr__(self):
        return "InsertTransformation('{}', {})".format(self.text, self.position)



from pathlib import Path

from PySide2.QtCore import QAbstractListModel


class FileSequence(QAbstractListModel):
    """A class used to represent a list of file objects"""

    def __init__(self, file_path_list):
        super().__init__()

        # self.files = [Path(file) for file in file_path_list if file.exists()]
        if file_path_list:
            self.files = [Path(file) for file in file_path_list]
        else:
            self.files = []

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.files)

    def data(self, QModelIndex, role=None):
        return self.files[QModelIndex.row()].name

    def __repr__(self):
        return str([str(file) for file in self.files])



from polyrename.transformation.transformation import Transformation


class PrefixTransformation(Transformation):
    schema = {
        "metadata": {"name": "Prefix", "description": "Prepend text to filename"},
        "options": [
            {
                "name": "Text",
                "description": "Text to prepend",
                "datatype": str,
                "required": True,
            }
        ],
    }

    def __init__(self, text):
        self.text = text

    def resolve(self, file_sequence):
        return_sequence = []
        for file in file_sequence:
            file_name = self.text + file.name
            file_path = file.parent / file_name
            return_sequence.append(file_path)

        return return_sequence

    def __repr__(self):
        return "PrefixTransformation('{}')".format(self.text)



from polyrename.transformation import (
    transformation,
    date_time,
    prefix,
    suffix,
    sequence,
    current_datetime,
    insert,
    remove_range,
    replace_extension,
    replace,
    regex_replace,
    music_tags,
)


# List of all available transformations
TRANSFORMATIONS = transformation.Transformation.__subclasses__()

TRANSFORMATIONS_BY_NAME = {t.schema["metadata"]["name"]: t for t in TRANSFORMATIONS}



from polyrename.transformation.transformation import Transformation


class ReplaceTransformation(Transformation):
    schema = {
        "metadata": {"name": "Replace", "description": "Replace text"},
        "options": [
            {
                "name": "Match Text",
                "description": "Text to match",
                "datatype": str,
                "required": True,
            },
            {
                "name": "Replace",
                "description": "Text to replace match with",
                "datatype": str,
                "required": False,
                "default_value": "",
            },
        ],
    }

    def __init__(self, match, replace):
        self.match = match
        self.replace = replace

    def resolve(self, file_sequence):
        return_sequence = []
        for file in file_sequence:

            file = file.parent / (
                file.stem.replace(self.match, self.replace) + file.suffix
            )

            return_sequence.append(file)

        return return_sequence

    def __repr__(self):
        return "ReplaceTransformation('{}'. '{}')".format(self.match, self.replace)



import signal
import sys

from PySide2.QtWidgets import QApplication

from polyrename.driver.gui.main_window import MainWindow


def main(files):
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    app = QApplication(sys.argv)

    window = MainWindow(files)

    sys.exit(app.exec_())



import logging

from PySide2.QtWidgets import (
    QGroupBox,
    QVBoxLayout,
    QListView,
    QAbstractItemView,
    QPushButton,
    QFileDialog,
)
from PySide2.QtGui import QStandardItem, QStandardItemModel

from polyrename.file_sequence import FileSequence


class FilePicker(QGroupBox):
    def __init__(self, files):
        super().__init__("File Picker")

        self.setLayout(QVBoxLayout())

        self.file_sequence = FileSequence([])

        self._initialize_file_list()

        self.file_sequence = FileSequence(files)
        self._update_file_picker_list([str(file) for file in files])

    def log_file_sequence_status(self):
        logging.debug("Current file sequence: {}".format(self.file_sequence))

    def clear_file_list(self):
        self.file_sequence = FileSequence([])
        self.file_list.model().clear()
        self.log_file_sequence_status()

    def _initialize_file_list(self):
        self.file_list = QListView()
        self.file_list.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.layout().addWidget(self.file_list)

        self.file_list.setModel(QStandardItemModel())

        select_files = QPushButton("Select Files")
        select_files.clicked.connect(self._select_files_listener)
        self.layout().addWidget(select_files)

    def _update_file_picker_list(self, file_names):
        model = self.file_list.model()

        # Update File Picker list from files[]
        model.clear()
        for f in range(self.file_sequence.rowCount()):
            item = QStandardItem()
            item.setText(file_names[f])
            model.appendRow(item)

        self.log_file_sequence_status()

    def _select_files_listener(self):
        """Handles selection of files to rename"""

        files = QFileDialog.getOpenFileNames(self, "Select Files", ".")
        self.file_sequence = FileSequence(files[0])

        self._update_file_picker_list(files[0])



from polyrename.transformation.transformation import Transformation


class RemoveRangeTransformation(Transformation):
    schema = {
        "metadata": {
            "name": "Remove Range",
            "description": "Removes range of text in file name",
        },
        "options": [
            {
                "name": "Start",
                "description": "Start position (inclusive)",
                "datatype": int,
                "required": True,
            },
            {
                "name": "Stop",
                "description": "Stop position (exclusive)",
                "datatype": int,
                "required": True,
            },
        ],
    }

    def __init__(self, start, stop):
        self.start = start if start >= 0 else 0
        self.stop = stop

    def resolve(self, file_sequence):
        return_sequence = []
        for file in file_sequence:

            # TODO: Make this clearer
            file_name_characters = [x for x in file.stem]
            for _ in range(self.stop - self.start):
                try:
                    del file_name_characters[self.start]
                except IndexError:
                    break

            file = file.parent / ("".join(file_name_characters) + file.suffix)

            return_sequence.append(file)

        return return_sequence

    def __repr__(self):
        return "RemoveRangeTransformation({}, {})".format(self.start, self.stop)



import logging

from PySide2.QtWidgets import QGroupBox, QVBoxLayout, QListView, QAbstractItemView
from PySide2.QtGui import QStandardItemModel, QStandardItem

from polyrename.transformation import TRANSFORMATIONS, TRANSFORMATIONS_BY_NAME


class TransformationLibrary(QGroupBox):
    def __init__(self, transformation_confirmation):
        super().__init__("Transformation Library")

        self.transformation_configuration = transformation_confirmation

        self.setLayout(QVBoxLayout())

        self._initialize_library()

    def _initialize_library(self):
        self.transformation_list = QListView()
        self.transformation_list.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.layout().addWidget(self.transformation_list)

        self.transformation_list.setModel(QStandardItemModel())

        self.transformation_list.clicked.connect(self.configure_transformation)

        # Populate transformations list
        for transformation in TRANSFORMATIONS:
            transformation_name = transformation.schema["metadata"]["name"]
            transformation_description = transformation.schema["metadata"][
                "description"
            ]

            transformation_element = QStandardItem()
            transformation_element.setText(transformation_name)
            transformation_element.setToolTip(transformation_description)

            self.transformation_list.model().appendRow(transformation_element)

    def configure_transformation(self, x):
        """Resolve selected transformation and pass it to transformation_configuration to display config"""

        model = self.transformation_list.model()
        selected_transformation = TRANSFORMATIONS_BY_NAME[model.data(x)]
        logging.debug("Selected transformation: {}".format(selected_transformation))

        self.transformation_configuration.swap_configuration(selected_transformation)



from polyrename.transformation.transformation import Transformation


class ReplaceExtensionTransformation(Transformation):
    schema = {
        "metadata": {
            "name": "Replace Extension",
            "description": "Replaces file extension with specified extension",
        },
        "options": [
            {
                "name": "Extension",
                "description": "Extension to use in replacement",
                "datatype": str,
                "required": False,
                "default_value": "",
            }
        ],
    }

    def __init__(self, extension):
        self.extension = extension

    def resolve(self, file_sequence):
        return_sequence = []
        for file in file_sequence:
            file_path = file.parent / (file.stem + self.extension)
            return_sequence.append(file_path)

        return return_sequence

    def __repr__(self):
        return "ReplaceExtension({})".format(self.extension)



import argparse
import logging
from pathlib import Path

from polyrename.driver import driver_gui, driver_dev


def _parse_args():
    parser = argparse.ArgumentParser(
        description="A cross-platform, bulk-file rename tool"
    )
    parser.add_argument("--verbose", "-v", action="count", default=0, help="Verbosity")
    parser.add_argument("-mode", choices=["gui", "dev"], default="gui", help="UI Mode")
    parser.add_argument("-files", nargs="+", default=[], help="Files to Rename")

    return parser.parse_args()


def main():
    args = _parse_args()

    if args.verbose == 0:
        logging.basicConfig(level=logging.WARNING)
    elif args.verbose == 1:
        logging.basicConfig(level=logging.INFO)
    elif args.verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)

    mode = args.mode
    files = [
        Path(file)
        for file in args.files
        if Path(file).exists() and Path(file).is_file()
    ]

    if mode == "gui":
        driver_gui.main(files)
    elif mode == "dev":
        driver_dev.main(files)


if __name__ == "__main__":
    main()



from polyrename.transformation.transformation import Transformation
from polyrename.transformation.utils.path_utils import insert_text_before_extension

import datetime


class CurrentDateTimeTransformation(Transformation):
    schema = {
        "metadata": {
            "name": "Current Datetime",
            "description": "Appends the current datetime accourding to a format string",
        },
        "options": [
            {
                "name": "Format String",
                "description": "Format string for datetime (see https://docs.python.org/3.7/library/datetime.html#strftime-and-strptime-behavior)",
                "datatype": str,
                "required": True,
            }
        ],
    }

    def __init__(self, format_string):
        self.format_string = format_string

    def resolve(self, file_sequence):
        datetime_object = datetime.datetime.now()
        resolved_datetime_string = datetime_object.strftime(self.format_string)

        return_sequence = []
        for file in file_sequence:
            file_path = insert_text_before_extension(file, resolved_datetime_string)
            return_sequence.append(file_path)

        return return_sequence

    def __repr__(self):
        return "CurrentDateTimeTransformation('{}')".format(self.format_string)



import datetime
from pathlib import Path

from polyrename.transformation import (
    prefix,
    suffix,
    date_time,
    sequence,
    current_datetime,
    insert,
    remove_range,
    replace_extension,
    replace,
    regex_replace,
)
from tests import TEST_SEQUENCE_01


def test_prefix():
    """Verify that prefix tranformation behaves as expected"""

    transformation = prefix.PrefixTransformation("HelloWorld")

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/HelloWorldexample.py"),
        Path("HelloWorldfile_without_extension"),
        Path("HelloWorldrelative.jpg"),
        Path("relative/path/HelloWorldextreme_extreme04.tar.gz"),
    ]


def test_suffix():
    """Verify that suffix tranformation behaves as expected"""

    # TODO: Determine how we want to handle things like 'example.tar.gz'
    transformation = suffix.SuffixTransformation("HelloWorld")

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/exampleHelloWorld.py"),
        Path("file_without_extensionHelloWorld"),
        Path("relativeHelloWorld.jpg"),
        Path("relative/path/extreme_extreme04.tarHelloWorld.gz"),
    ]


def test_date_time():
    """Verify that date transformation behaves as expected"""

    transformation = date_time.DateTimeTransformation(
        2012, 1, 3, 0, 0, 0, 0, " (%Y-%m-%d)"
    )

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/example (2012-01-03).py"),
        Path("file_without_extension (2012-01-03)"),
        Path("relative (2012-01-03).jpg"),
        Path("relative/path/extreme_extreme04.tar (2012-01-03).gz"),
    ]


def test_sequence():
    """Verify that sequence transformation behaves as expected"""

    transformation = sequence.SequenceTransformation(10, 22, "a", 4)

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/exampleaa10.py"),
        Path("file_without_extensionaa32"),
        Path("relativeaa54.jpg"),
        Path("relative/path/extreme_extreme04.taraa76.gz"),
    ]


def test_current_datetime():
    """Verify that current datetime transformation behaves as expected"""

    format_string = "%Y-%m-%d %H:%M:%S"

    transformation = current_datetime.CurrentDateTimeTransformation(format_string)
    resolved = transformation.resolve(TEST_SEQUENCE_01)

    now = datetime.datetime.now()

    # Make sure that all datetimes are the same
    datetime_portions = [x.stem[-19:] for x in resolved]

    assert len(set(datetime_portions)) == 1

    then = datetime.datetime.strptime(datetime_portions[0], format_string)
    elapsed = now - then

    assert elapsed.seconds < 240


def test_insert():
    """Verify that insert transformation behaves as expected"""

    transformation = insert.InsertTransformation("hello", 7)

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/examplehello.py"),
        Path("file_wihellothout_extension"),
        Path("relativhelloe.jpg"),
        Path("relative/path/extremehello_extreme04.tar.gz"),
    ]


def test_remove_range():
    """Verify that remove range transformation behaves as expected"""

    transformation = remove_range.RemoveRangeTransformation(-4, 0)

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/example.py"),
        Path("file_without_extension"),
        Path("relative.jpg"),
        Path("relative/path/extreme_extreme04.tar.gz"),
    ]

    transformation = remove_range.RemoveRangeTransformation(0, 2)

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/ample.py"),
        Path("le_without_extension"),
        Path("lative.jpg"),
        Path("relative/path/treme_extreme04.tar.gz"),
    ]

    transformation = remove_range.RemoveRangeTransformation(0, 100)

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/.py"),
        Path(""),
        Path(".jpg"),
        Path("relative/path/.gz"),
    ]

    transformation = remove_range.RemoveRangeTransformation(5, 32)

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/examp.py"),
        Path("file_"),
        Path("relat.jpg"),
        Path("relative/path/extre.gz"),
    ]

    transformation = remove_range.RemoveRangeTransformation(100, 500)

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/example.py"),
        Path("file_without_extension"),
        Path("relative.jpg"),
        Path("relative/path/extreme_extreme04.tar.gz"),
    ]


def test_replace_extension():
    """Verify that replace extension transformation behaves as expected"""

    transformation = replace_extension.ReplaceExtensionTransformation(".txt")

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/example.txt"),
        Path("file_without_extension.txt"),
        Path("relative.txt"),
        Path("relative/path/extreme_extreme04.tar.txt"),
    ]


def test_replace():
    """Verify that replace transformation behaves as expected"""

    transformation = replace.ReplaceTransformation("ex", "HELLO")

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/HELLOample.py"),
        Path("file_without_HELLOtension"),
        Path("relative.jpg"),
        Path("relative/path/HELLOtreme_HELLOtreme04.tar.gz"),
    ]


def test_regex_replace():
    """Verify that regex replace transformation behaves as expected"""

    transformation = regex_replace.RegexReplaceTransformation("[_w]", "HelloWorld")

    assert transformation.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/example.py"),
        Path("fileHelloWorldHelloWorldithoutHelloWorldextension"),
        Path("relative.jpg"),
        Path("relative/path/extremeHelloWorldextreme04.tar.gz"),
    ]


# TODO: Implement Test
# def test_music_transformation():



from polyrename.transformation.transformation import Transformation
from polyrename.transformation.utils.path_utils import insert_text_before_extension


class SequenceTransformation(Transformation):
    schema = {
        "metadata": {"name": "Sequence", "description": "Append numeric sequence"},
        "options": [
            {
                "name": "Start",
                "description": "Initial value of sequence",
                "datatype": int,
                "required": True,
                "default_value": 1,
            },
            {
                "name": "Step",
                "description": "Value to increment sequence by",
                "datatype": int,
                "required": True,
                "default_value": 1,
            },
            {
                "name": "Pad Character",
                "description": "Character to pad sequence number by",
                "datatype": str,
                "required": True,
                "default_value": "0",
            },
            {
                "name": "Pad length",
                "description": "Number of characters to pad sequence to",
                "datatype": int,
                "required": True,
                "default_value": 1,
            },
        ],
    }

    def __init__(self, start, step, pad_char, pad_len):
        self.start = start
        self.step = step
        self.pad_char = pad_char
        self.pad_len = pad_len

    def resolve(self, file_sequence):
        return_sequence = []
        sequence_value = self.start
        for file in file_sequence:
            file_path = insert_text_before_extension(
                file, str(sequence_value).rjust(self.pad_len, self.pad_char)
            )
            return_sequence.append(file_path)

            sequence_value += self.step

        return return_sequence

    def __repr__(self):
        return "SequenceTransformation({}, {}, '{}', {})".format(
            self.start, self.step, self.pad_char, self.pad_len
        )



import pytag

from polyrename.transformation.transformation import Transformation
from polyrename.transformation.utils.path_utils import insert_text_before_extension


class MusicTagsTransformation(Transformation):
    schema = {
        "metadata": {
            "name": "Music Tags",
            "description": "Append a formatted string that pulls data from music metadata",
        },
        "options": [
            {
                "name": "Format String",
                "description": "Format string for text to append",
                "datatype": str,
                "required": True,
            }
        ],
    }

    format_map = {"%t": "title", "%n": "tracknumber", "%a": "artist"}

    forbidden_characters = [r"<", r">", r":", r'"', r"/", r"\\", r"|", r"?", r"*"]

    def __init__(self, format_string):
        self.format_string = format_string

    def sanitize_tag(self, tag):
        for character in self.forbidden_characters:
            tag = tag.replace(character, "")

        tag = tag.replace("\x00", "")

        return tag

    def resolve(self, file_sequence):
        return_sequence = []

        for file in file_sequence:

            # Skip if not a supported music file
            try:
                music_file = pytag.Audio(str(file))
            except:
                return_sequence.append(file)
                continue

            formatted_string = self.format_string
            for formatter, replacement in self.format_map.items():

                try:
                    tag_value = music_file.get_tags()[replacement]
                except KeyError as e:
                    print(e)
                    formatted_string = formatted_string.replace(formatter, "")
                    continue

                tag_value = self.sanitize_tag(tag_value)

                formatted_string = formatted_string.replace(formatter, tag_value)

            return_sequence.append(insert_text_before_extension(file, formatted_string))

        return return_sequence

    def __repr__(self):
        return "MusicTagsTransformation('{}')".format(self.format_string)



from PySide2.QtCore import QAbstractListModel


class Pipeline(QAbstractListModel):
    def __init__(self):
        super().__init__()

        self.transformations = []

    def add_transformation(self, transformation):
        self.transformations.append(transformation)

    def remove_transformation(self, index):
        del self.transformations[index]

    def move_transformation_up(self, index):
        if index > 0:
            temp = self.transformations[index]
            self.transformations[index] = self.transformations[index - 1]
            self.transformations[index - 1] = temp

    def move_transformation_down(self, index):
        if index < len(self.transformations) - 1:
            temp = self.transformations[index]
            self.transformations[index] = self.transformations[index + 1]
            self.transformations[index + 1] = temp

    def resolve(self, file_sequence):
        for transformation in self.transformations:
            file_sequence = transformation.resolve(file_sequence)

        return file_sequence

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.transformations)

    def data(self, QModelIndex, role=None):
        return self.transformations[QModelIndex]

    def __str__(self):
        return "Pipeline({})".format(self.transformations)



from pathlib import Path

from PySide2.QtWidgets import QApplication

from polyrename.transformation import prefix, suffix, pipeline
from tests import TEST_SEQUENCE_01


def test_pipeline():
    """Ensures that pipelines successfully string together transformations"""

    app = QApplication([])

    prefix_transformation = prefix.PrefixTransformation("Hello_")
    suffix_transformation = suffix.SuffixTransformation("_World")

    prefix_suffix_pipeline = pipeline.Pipeline()
    prefix_suffix_pipeline.add_transformation(prefix_transformation)
    prefix_suffix_pipeline.add_transformation(suffix_transformation)

    assert prefix_suffix_pipeline.resolve(TEST_SEQUENCE_01) == [
        Path("/home/test/Hello_example_World.py"),
        Path("Hello_file_without_extension_World"),
        Path("Hello_relative_World.jpg"),
        Path("relative/path/Hello_extreme_extreme04.tar_World.gz"),
    ]



from polyrename.transformation.transformation import Transformation
from polyrename.transformation.utils.path_utils import insert_text_before_extension

import datetime


class DateTimeTransformation(Transformation):
    schema = {
        "metadata": {
            "name": "Datetime",
            "description": "Appends a formatted datetime object",
        },
        "options": [
            {
                "name": "Year",
                "description": "Date year",
                "datatype": int,
                "required": False,
                "default_value": 2000,
            },
            {
                "name": "Month",
                "description": "Date month",
                "datatype": int,
                "required": False,
                "default_value": 1,
            },
            {
                "name": "Day",
                "description": "Date day",
                "datatype": int,
                "required": False,
                "default_value": 1,
            },
            {
                "name": "Hour",
                "description": "Time hour",
                "datatype": int,
                "required": False,
                "default_value": 0,
            },
            {
                "name": "Minute",
                "description": "Time minute",
                "datatype": int,
                "required": False,
                "default_value": 0,
            },
            {
                "name": "Second",
                "description": "Time second",
                "datatype": int,
                "required": False,
                "default_value": 0,
            },
            {
                "name": "Microsecond",
                "description": "Time microsecond",
                "datatype": int,
                "required": False,
                "default_value": 0,
            },
            {
                "name": "Format String",
                "description": "Format string for datetime (see https://docs.python.org/3.7/library/datetime.html#strftime-and-strptime-behavior)",
                "datatype": str,
                "required": True,
            },
        ],
    }

    def __init__(
        self, year, month, day, hour, minute, second, microsecond, format_string
    ):
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.microsecond = microsecond
        self.format_string = format_string

    def resolve(self, file_sequence):

        datetime_object = datetime.datetime(
            *(
                self.year,
                self.month,
                self.day,
                self.hour,
                self.minute,
                self.second,
                self.microsecond,
            )
        )
        resolved_datetime_string = datetime_object.strftime(self.format_string)

        return_sequence = []
        for file in file_sequence:
            file_path = insert_text_before_extension(file, resolved_datetime_string)
            return_sequence.append(file_path)

        return return_sequence

    def __repr__(self):
        return "DateTimeTransformation({}, {}, {}, {}, {}, {}, {}, '{}')".format(
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
            self.microsecond,
            self.format_string,
        )



from pathlib import Path
from setuptools import setup, find_packages

HERE = Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name='PolyRename',
    version='1.0.0',
    license='GPLv3',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    author='Andrew Simmons, Ethan Smith',
    author_email='agsimmons0@gmail.com',
    description='A cross-platform, bulk-file rename tool',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/agsimmons/PolyRename',
    project_urls={
        'Documentation': 'https://polyrename.readthedocs.io/',
        'Bug Tracker': 'https://github.com/agsimmons/PolyRename/issues/',
    },

    packages=find_packages(),
    include_package_data=True,
    package_data={
        'gui': ['images/*.png']
    },
    zip_safe=False,

    entry_points={
        'console_scripts': ['polyrename=polyrename.__main__:main']
    },

    install_requires=[
        'PySide2>=5.11.0',
        'pytag>=0.1.5',
    ]
)



from polyrename.transformation import TRANSFORMATIONS


VALID_DATATYPES = [str, int, bool]


def test_schemas():
    """Validate that each transformation's schema complies with the schema specification"""

    for transformation in TRANSFORMATIONS:

        # Test that transformation has a schema
        schema = transformation.schema

        # Test metadata block
        assert "metadata" in schema
        # Name
        assert "name" in schema["metadata"]
        assert isinstance(schema["metadata"]["name"], str)
        assert schema["metadata"]["name"] != ""
        # Description
        assert "description" in schema["metadata"]
        assert isinstance(schema["metadata"]["description"], str)
        assert schema["metadata"]["description"] != ""

        # Test options block
        assert "options" in schema
        assert isinstance(schema["options"], list)
        assert len(schema["options"]) > 0
        for option in schema["options"]:
            # Name
            assert "name" in option
            assert isinstance(option["name"], str)
            assert option["name"] != ""
            # Description
            assert "description" in option
            assert isinstance(option["description"], str)
            assert option["description"] != ""
            # Data Type
            assert "datatype" in option
            assert option["datatype"] in VALID_DATATYPES
            # Required
            assert "required" in option
            assert isinstance(option["required"], bool)
            if not option["required"]:
                assert "default_value" in option
                assert isinstance(option["default_value"], option["datatype"])


def test_unique_names():
    """Validate that names are unique among transformations"""

    names = []
    for transformation in TRANSFORMATIONS:
        names.append(transformation.schema["metadata"]["name"])

    assert len(names) == len(set(names))



import logging
import shutil

from PySide2.QtWidgets import (
    QGroupBox,
    QVBoxLayout,
    QPushButton,
    QListView,
    QMessageBox,
    QHBoxLayout,
    QAbstractItemView,
    QWidget,
)
from PySide2.QtGui import QStandardItem, QStandardItemModel

from polyrename.transformation.pipeline import Pipeline


class PipelineEditor(QGroupBox):
    def __init__(self, file_picker):
        super().__init__("Pipeline Editor")

        self.setLayout(QVBoxLayout())

        self.file_picker = file_picker

        self.pipeline = Pipeline()
        self.pipelineView = QListView()
        self.pipelineView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.pipelineView.setModel(QStandardItemModel())

        self.layout().addWidget(self.pipelineView)

        # === BUTTON CONTAINER ===
        button_rows = QWidget()
        self.layout().addWidget(button_rows)
        button_rows_layout = QVBoxLayout(button_rows)
        button_rows_layout.setContentsMargins(0, 0, 0, 0)

        # === ROW 1 ===
        row_1 = QWidget()
        button_rows_layout.addWidget(row_1)
        row_1_layout = QHBoxLayout(row_1)
        row_1_layout.setContentsMargins(0, 0, 0, 0)

        self.moveUpButton = QPushButton("Move Up")
        row_1_layout.addWidget(self.moveUpButton)
        self.moveUpButton.clicked.connect(self._move_up_listener)

        self.moveDownButton = QPushButton("Move Down")
        row_1_layout.addWidget(self.moveDownButton)
        self.moveDownButton.clicked.connect(self._move_down_listener)

        self.deleteButton = QPushButton("Delete")
        row_1_layout.addWidget(self.deleteButton)
        self.deleteButton.clicked.connect(self._delete_listener)

        # === ROW 2 ===
        row_2 = QWidget()
        button_rows_layout.addWidget(row_2)
        row_2_layout = QHBoxLayout(row_2)
        row_2_layout.setContentsMargins(0, 0, 0, 0)

        self.applyButton = QPushButton("Apply Pipeline")
        row_2_layout.addWidget(self.applyButton)
        self.applyButton.clicked.connect(self._apply_pipeline_listener)

        self.update_pipeline_view()

    def update_pipeline_view(self):
        logging.debug(self.pipeline)
        model = self.pipelineView.model()
        model.clear()

        for t in range(self.pipeline.rowCount()):
            item = QStandardItem()
            item.setText(repr(self.pipeline.data(t)))
            model.appendRow(item)

    def _modify_transformation_listener(self):
        # TODO
        raise NotImplementedError

    def _move_up_listener(self):
        try:
            to_move = self.pipelineView.selectionModel().selectedIndexes()[0].row()
            if to_move < 1:
                return
            self.pipeline.move_transformation_up(to_move)
            self.update_pipeline_view()
            self.pipelineView.setCurrentIndex(
                self.pipelineView.model().index(to_move - 1, 0)
            )
        except IndexError:
            return

    def _move_down_listener(self):
        try:
            to_move = self.pipelineView.selectionModel().selectedIndexes()[0].row()
            if to_move > self.pipelineView.model().rowCount() - 2:
                return
            self.pipeline.move_transformation_down(to_move)
            self.update_pipeline_view()
            self.pipelineView.setCurrentIndex(
                self.pipelineView.model().index(to_move + 1, 0)
            )
        except IndexError:
            return

    def _delete_listener(self):
        try:
            to_delete = self.pipelineView.selectionModel().selectedIndexes()[0].row()
            self.pipeline.remove_transformation(to_delete)
            self.update_pipeline_view()
            if to_delete > self.pipelineView.model().rowCount() - 1:
                self.pipelineView.setCurrentIndex(
                    self.pipelineView.model().index(to_delete - 1, 0)
                )
            else:
                self.pipelineView.setCurrentIndex(
                    self.pipelineView.model().index(to_delete, 0)
                )
        except IndexError:
            return

    def _apply_pipeline_listener(self):
        # Check that at least one transformation has been added to the pipeline
        if self.pipeline.rowCount() == 0:
            no_transformations_messagebox = QMessageBox(self)
            no_transformations_messagebox.setText("ERROR: No transformations selected")
            no_transformations_messagebox.exec_()
            return

        file_sequence = self.file_picker.file_sequence.files

        # Check that at least one file has been added to the file sequence
        if len(file_sequence) == 0:
            no_files_messagebox = QMessageBox(self)
            no_files_messagebox.setText("ERROR: No files selected")
            no_files_messagebox.exec_()
            return

        transformed_sequence = self.pipeline.resolve(file_sequence)

        before_after = list(zip(file_sequence, transformed_sequence))

        preview_text_lines = []
        for rename in before_after:
            preview_text_lines.append("{} -> {}".format(rename[0].name, rename[1].name))
        preview_text = "\n".join(preview_text_lines)

        confirmation = QMessageBox(self)
        confirmation.setText("Are you sure you want to apply the pipeline?")
        confirmation.setDetailedText(preview_text)
        confirmation.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        confirmation.setDefaultButton(QMessageBox.No)
        ret = confirmation.exec_()

        if ret == int(QMessageBox.Yes):
            for rename in before_after:
                from_path = rename[0]
                to_path = rename[1]
                
                shutil.move(str(from_path), str(to_path))
            self.file_picker.clear_file_list()



import logging

from PySide2.QtWidgets import (
    QGroupBox,
    QVBoxLayout,
    QScrollArea,
    QWidget,
    QFormLayout,
    QLabel,
    QTextEdit,
    QPushButton,
    QMessageBox,
)
from PySide2.QtGui import QFontMetrics


class TransformationConfiguration(QGroupBox):
    def __init__(self, pipeline_editor):
        super().__init__("Transformation Config")

        self.pipeline_editor = pipeline_editor

        self.selected_transformation = None

        self.setLayout(QVBoxLayout())

        self._initialize_text()
        self._initialize_scroll_area()
        self._initialize_add_button()

    def _initialize_text(self):
        metrics = QFontMetrics(QTextEdit().font())
        self.text_line_height = metrics.lineSpacing() * 2

    def _initialize_scroll_area(self):
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)

        self.config_form = QFormLayout()

        form_widget = QWidget()
        form_widget.setLayout(self.config_form)

        scroll_area.setWidget(form_widget)

        self.layout().addWidget(scroll_area)

    def _initialize_add_button(self):
        add_button = QPushButton("Add to Pipeline")
        add_button.clicked.connect(self.add_configured_transformation_to_pipeline)

        self.layout().addWidget(add_button)

    def swap_configuration(self, selected_transformation):
        """Swap config form with that of selected_transformation"""

        self.selected_transformation = selected_transformation

        # Destroy existing form rows
        for i in reversed(range(self.config_form.count())):
            self.config_form.itemAt(i).widget().setParent(None)

        # Generate configuration for selected Transformation
        for option in selected_transformation.schema["options"]:

            label = QLabel(option["name"])
            label.setToolTip(option["description"])

            if "default_value" in option and option["required"] is True:
                field = QTextEdit(str(option["default_value"]))
            else:
                field = QTextEdit("")
            field.setFixedHeight(self.text_line_height)

            self.config_form.addRow(label, field)

    def add_configured_transformation_to_pipeline(self):
        try:
            options = self.selected_transformation.schema["options"]
        except AttributeError:
            # If no transformation is selected yet
            return

        # Get form options
        form_options = []
        for i in range(self.config_form.count()):

            # Skip labels in the form
            if i % 2 == 0:
                continue

            configurated_field = self.config_form.itemAt(i).widget()
            form_options.append(configurated_field.toPlainText())
        logging.debug("Selected options: {}".format(form_options))

        option_types = [option["datatype"] for option in options]
        logging.debug("Option types: {}".format(option_types))

        # Fill in default values if an option has one and the user entered nothing into a field
        for i in range(len(form_options)):
            if (
                options[i]["required"] is False
                and "default_value" in options[i]
                and len(form_options[i]) == 0
            ):
                form_options[i] = options[i]["default_value"]
        logging.debug("Defaults filled in: {}".format(form_options))

        # Error if a required option is still not filled
        for i in range(len(form_options)):
            if options[i]["required"] and len(form_options[i]) == 0:
                missing_required_field_messagebox = QMessageBox(self)
                missing_required_field_messagebox.setText(
                    "ERROR: Field '{}' is required".format(options[i]["name"])
                )
                missing_required_field_messagebox.exec_()
                return

        # Convert form options to correct datatype
        for i in range(len(form_options)):
            try:
                form_options[i] = option_types[i](form_options[i])
            except ValueError:
                invalid_datatype_messagebox = QMessageBox(self)
                invalid_datatype_messagebox.setText(
                    "ERROR! Invalid input datatype for field: {}".format(
                        options[i]["name"]
                    )
                )
                invalid_datatype_messagebox.exec_()
                return
        logging.debug("Casted options: {}".format(form_options))

        configured_transformation = self.selected_transformation(*form_options)
        logging.debug("Configured Transformation: {}".format(configured_transformation))

        self.pipeline_editor.pipeline.add_transformation(configured_transformation)
        self.pipeline_editor.update_pipeline_view()
